/*
 *
 *  Copyright Peter Boyle and Glasgow University, 2000.
 *  
 *  It is provided under the GNU pubic License V2
 *  It is provided as is and is not guaranteed fit for any purpose.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "processor.h"
#include "powerpc.h"


struct processor *create_bgq ()
{
  struct regs *iregs=new regs;
  struct regs *fregs=new regs;
  struct processor *proc = new processor;
  struct pipeline  *pipes = new pipeline[2]; /*Hardwire 2 pipes for now!!*/
  struct instructionset *iset = new instructionset;

  int i;

  iregs->regfile=32;
  fregs->regfile=32;
  iregs->fmt="%%r%d";
  fregs->fmt="%d";
  for(i=0;i<32;i++){
    iregs->allocated[i] = 0;
    fregs->allocated[i] = 0;
    iregs->mnemonic[i] = NULL;
    fregs->mnemonic[i] = NULL;
  }
 
  proc->iregs = iregs;
  proc->fregs = fregs;
  proc->pipes = pipes;
  proc->nsimd = 2;/*Number of COMPLEX elements in a SIMD operation*/
  proc->iset = iset;
  proc->FP_size = 8;
  proc->FSP_size = 4;
  proc->I_size = 8;
  proc->IS_size= 4;
  proc->usenops = 0;
  proc->delayslot = 0;
  proc->RegWindows = 0;
  proc->Bias = 0;
  proc->lmq_space=20;
  proc->l1_locking=1;
  proc->load_reorder=0;
  proc->cross_negative=1;

   /*{... ABI definition ...*/
  proc->ABI_argbase = 3;  
    /*Allocate dedicated registers*/
  iregs->allocated[0] = 1; /*Don't use GPR 0 since it can have anomolous behaviour*/
                           /* when used in an effective address if used by accident*/
  iregs->allocated[1] = 1;
  iregs->allocated[2] = 1;
  proc->StackPointer = 1 ;
  iregs->mnemonic[0] = "GPR0";
  iregs->mnemonic[1] = "SP";
  iregs->mnemonic[2] = "GPR2";
  proc->ReturnPointer =-1; /*Return is in the link register, not a GPR*/

    /*ABI defines a number of callee save registers*/
  proc->Isave=0;
  for ( i=2; i<=31 ; i ++){
    proc->Isave= proc->Isave | (1<<i);
  }
  proc->Isave = proc->Isave | 1;
  proc->Fsave=0;
  for ( i=0; i<=31 ; i ++){
    proc->Fsave= proc->Fsave | (1<<i);
  }
  /*...END ABI}*/

  proc->CacheLine = 64; /*64 byte cache line*/

  /*Actually issue width is 2, but just pretend*/
  proc->npipes=2;
  for(i=0;i<N_PIPE_TYPES;i++){
    proc->switchargs[i] = 0; /*Wrote this with PPC in mind so arg order is always native*/
  }
  proc->switchargs[SIMDHUMMERPIPE] = 1; 

  pipes[0].issuegroups = FADDMASK | FMULMASK | FMACMASK |FMOVMASK |
                         SIMDHUMMERMASK | SIMDPERMUTEMASK;
  pipes[1].issuegroups = IALUMASK | BRCHMASK | LOADMASK | 
                         STORMASK | FLODMASK | CACHMASK | FSTRMASK ;
  int qpx_lat = 8;
  proc->latency[SIMDHUMMERPIPE] = qpx_lat;
  proc->latency[FMACPIPE] = 8;
  proc->latency[FADDPIPE] = 6;
  proc->latency[FMULPIPE] = 8;
  proc->latency[FMOVPIPE] = 3;
  proc->latency[FLODPIPE] = 3;
  proc->latency[FSTRPIPE] = 2;
  proc->latency[IALUPIPE] = 2;
  proc->latency[LOADPIPE] = 2;
  proc->latency[STORPIPE] = 2;
  proc->latency[SIMDPERMUTEPIPE] = 6;

  set_scheduler_bgq();
  set_scheduler_weird();

  create_bgq_instructions(proc);
  return(proc);
}

void create_bgq_instructions( struct processor *proc)
{
  int i,j;

  struct instruction *fmacs = new instruction [nfmacs];
  struct instruction *fadds = new instruction [nfadds];
  struct instruction *fmuls = new instruction [nfmuls];
  struct instruction *fmovs = new instruction [nfmovs];
  struct instruction *fsaves= new instruction [nfsaves];
  struct instruction *loads = new instruction [nloads];
  struct instruction *floads= new instruction [nfloads];
  struct instruction *cache = new instruction [ncache];
  struct instruction *adds  = new instruction [nadds];
  struct instruction *branch   = new instruction [nbrch];
  struct instruction *directive= new instruction [ndirective];
  struct instruction *stores   = new instruction [nstore];
  struct instruction *hummers  = new instruction [nsimdhummer];
  struct instruction *permutes = new instruction [nsimdpermute];
  struct instruction *weird    = new instruction [nweirdops];

  /*Define the FMACS*/
  fmacs[FMADD].mnemonic = "fmadd ";
  fmacs[FMSUB].mnemonic = "fmsub ";
  fmacs[FNMADD].mnemonic ="fnmadd";
  fmacs[FNMSUB].mnemonic ="fnmsub";
  for (i=0;i<nfmacs;i++){
    fmacs[i].narg  = 4;
    for(j=0;j<fmacs[i].narg;j++){
      fmacs[i].argtypes[j] = Fregs;
      fmacs[i].inout[j] = inarg;
    }
    fmacs[i].inout[0] = outarg;
    fmacs[i].pipe  = FMACPIPE;
    fmacs[i].code = i;
  }

  /*Define the permute pipe*/
  permutes[VPERM].mnemonic = "qvfperm ";
  permutes[VPERM].narg     = 4;
  permutes[VPERMIMM].mnemonic = "qvgpci ";
  permutes[VPERMIMM].narg     = 2;

  for(int i=0;i<nsimdpermute;i++){
  for(j=0;j<permutes[i].narg;j++){
    permutes[i].argtypes[j] = Fregs;
    permutes[i].inout[j] = inarg;
    permutes[i].code     = i;
    permutes[i].inout[0] = outarg;
  }}
  permutes[VPERMIMM].argtypes[1] = SIMM;

  /*Define the hummer*/
  hummers[VMADD].mnemonic = "qvfmadd ";
  hummers[VMSUB].mnemonic = "qvfmsub ";
  hummers[VNMADD].mnemonic ="qvfnmadd ";
  hummers[VNMSUB].mnemonic ="qvfnmsub ";
  hummers[VMADD_RR_RI].mnemonic ="qvfxmadd ";
  hummers[VMADD_MII_IR].mnemonic ="qvfxxnpmadd ";
  hummers[VMADD_II_MIR].mnemonic ="qvfxxcpnmadd ";

  hummers[VMUL].mnemonic = "qvfmul ";
  hummers[VADD].mnemonic = "qvfadd ";
  hummers[VSUB].mnemonic = "qvfsub ";
  hummers[VMUL_RR_RI].mnemonic ="qvfxmul ";/*Not implemented*/
  hummers[VNEG].mnemonic = "qvfneg ";
  for (i=0;i<nsimdhummer;i++){
    hummers[i].narg  = 4;
    for(j=0;j<hummers[i].narg;j++){
      hummers[i].argtypes[j] = Fregs;
      hummers[i].inout[j] = inarg;
    }
    hummers[i].inout[0] = outarg;
    hummers[i].pipe  = SIMDHUMMERPIPE;
    hummers[i].code = i;
  }
  hummers[VMUL].narg = 3;
  hummers[VADD].narg = 3;
  hummers[VSUB].narg = 3;
  hummers[VMUL_RR_RI].narg = 3;
  hummers[VNEG].narg = 2;

  /*Define the FADDS*/  
  fadds[FADD].mnemonic = "fadd  ";
  fadds[FADD].mnemonic = "fadd  ";
  fadds[FSUB].mnemonic = "fsub  ";
  for (i=0;i<nfadds;i++){
    fadds[i].narg = 3;
    for(j=0;j<fadds[i].narg;j++){
      fadds[i].argtypes[j] = Fregs;
      fadds[i].inout[j] = inarg;
    }
    fadds[i].inout[0] = outarg;
    fadds[i].pipe  = FADDPIPE;
    fadds[i].code = i;
  }

  /*Define the FMOV*/  
  fmovs[FMOV].mnemonic = "fmr  ";
  fmovs[FNEG].mnemonic = "fneg  ";
  for (i=0;i<nfmovs;i++){
    fmovs[i].narg = 2;
    for(j=0;j<fmacs[i].narg;j++){
      fmovs[i].argtypes[j] = Fregs;
      fmovs[i].inout[j] = inarg;
    }
    fmovs[i].inout[0] = outarg;
    fmovs[i].pipe  = FMOVPIPE;
    fmovs[i].code = i;
  }

  /*Define FMUL*/
  fmuls[FMUL].mnemonic = "fmul  ";
  fmuls[FMUL].narg  = 3 ;
  for(j=0;j<fmuls[FMUL].narg;j++){
    fmuls[FMUL].argtypes[j] = Fregs;
    fmuls[FMUL].inout[j] = inarg;
  }
  fmuls[FMUL].inout[0] = outarg;
  fmuls[FMUL].pipe  = FMULPIPE;
  fmuls[FMUL].code = FMUL;


  /*Define loads*/
  loads[ILD_EA].mnemonic = "ld  ";
  loads[LOAD_REG_EA].mnemonic = "ld  ";
  loads[LOAD_ADDR].mnemonic = "la  ";
  loads[LOAD_IMM].mnemonic = "li  ";
  loads[ILD_EA_SHORT].mnemonic = "lwz  ";
  for (i=0;i<nloads;i++){
    loads[i].narg = 3;
    loads[i].argtypes[0] = Iregs;
    loads[i].argtypes[1] = SIMM;
    loads[i].argtypes[2] = Iregs;
    loads[i].inout[0] = outarg;
    loads[i].inout[1] = inarg;
    loads[i].inout[2] = inarg;
    loads[i].pipe  = LOADPIPE;
    loads[i].code  = i;
  }
  loads[LOAD_IMM].narg = 2;

  /*Define STORE - really the register save and restore*/
  /*Since these must be full register width we abstract them*/
  /* differently from the floating and integer which can be a lower precision*/
  stores[SAVE_REG_EA].mnemonic = "std  ";
  stores[IST_EA].mnemonic = "std  ";
  for (i=0;i<nstore;i++){
    stores[i].narg = 3;
    stores[i].argtypes[0] = Iregs;
    stores[i].argtypes[1] = SIMM;
    stores[i].argtypes[2] = Iregs;
    stores[i].inout[0] = inarg;
    stores[i].inout[1] = inarg;
    stores[i].inout[2] = inarg;
    stores[i].pipe  = STORPIPE;
    stores[i].code  = i;
  }
 
  /*Define cache control codes*/
  cache[PREF_IMM].mnemonic = "dcbt "; /*Pre-fetch*/
  cache[WRITE_HINT].mnemonic = "dcbz "; /*Write hint - destructive*/
  cache[FLUSH].mnemonic = "dcbf ";
  cache[TOUCHLOCKSET].mnemonic = "dcbtls ";
  cache[TOUCHLOCKCLEAR].mnemonic = "dcblc ";
  cache[TOUCHL2].mnemonic = "dcbt ";
  cache[COUNTER_HINT].mnemonic = "mtctr ";
  for (i=0;i<ncache;i++){
    cache[i].narg = 1;
    cache[i].argtypes[0] = Iregs;
    cache[i].inout[0]  = inarg;
    cache[i].pipe  = CACHPIPE;
    cache[i].code  = i;
  }
  cache[PREF_IMM].narg = 2;
  cache[PREF_IMM].argtypes[0] = Iregs;
  cache[PREF_IMM].argtypes[1] = Iregs;
  cache[PREF_IMM].inout[1] = inarg;
  cache[TOUCHLOCKSET].narg = 2;
  cache[TOUCHLOCKSET].argtypes[0] = Iregs;
  cache[TOUCHLOCKSET].argtypes[1] = Iregs;
  cache[TOUCHLOCKSET].inout[1] = inarg;
  cache[TOUCHLOCKCLEAR].narg = 2;
  cache[TOUCHLOCKCLEAR].argtypes[0] = Iregs;
  cache[TOUCHLOCKCLEAR].argtypes[1] = Iregs;
  cache[TOUCHLOCKCLEAR].inout[1] = inarg;
  cache[TOUCHL2].narg = 2;
  cache[TOUCHL2].argtypes[0] = Iregs;
  cache[TOUCHL2].argtypes[1] = Iregs;
  cache[TOUCHL2].inout[1] = inarg;
  cache[FLUSH].narg = 2;
  cache[FLUSH].argtypes[0] = Iregs;
  cache[FLUSH].argtypes[1] = Iregs;
  cache[FLUSH].inout[1] = inarg;
  cache[WRITE_HINT].narg = 2;
  cache[WRITE_HINT].argtypes[0] = Iregs;
  cache[WRITE_HINT].argtypes[1] = Iregs;
  cache[WRITE_HINT].inout[1] = inarg;

  /*Floating point loads - does the APU interrupt the FPU arithmetic stream*/
  floads[FLOAD_EA].mnemonic = "lfd  "  ;
  floads[FLOAD_REG_EA].mnemonic = "lfd  " ;
  floads[FS_LOAD_EA].mnemonic = "lfs  "  ;
  floads[CLOAD_INDEXED].mnemonic = "qvlfdx  " ;
  floads[CS_LOAD_INDEXED].mnemonic = "qvlfsx  " ;
  for (i=0;i<nfloads;i++){
    floads[i].narg = 3;
    floads[i].argtypes[0] = Fregs;
    floads[i].argtypes[1] = SIMM;
    floads[i].argtypes[2] = Iregs;

    floads[i].inout[0]  = outarg;
    floads[i].inout[1]  = inarg;
    floads[i].inout[2]  = inarg;

    floads[i].pipe  = FLODPIPE;
    floads[i].code  = i;
  }
  floads[CLOAD_INDEXED].argtypes[1]=Iregs;
  floads[VLOAD_INDEXED].argtypes[1]=Iregs;
  floads[CS_LOAD_INDEXED].argtypes[1]=Iregs;

  fsaves[FSAVE_EA].mnemonic = "stfd " ;
  fsaves[FSAVE_REG_EA].mnemonic = "stfd " ;
  fsaves[CSAVE_INDEXED].mnemonic = "qvstfdx " ;
  fsaves[FS_SAVE_EA].mnemonic = "stfs " ;
  fsaves[CS_SAVE_INDEXED].mnemonic = "qvstfsx " ;
  for (i=0;i<nfsaves;i++){
    fsaves[i].narg = 3;
    fsaves[i].argtypes[0] = Fregs;
    fsaves[i].argtypes[1] = SIMM;
    fsaves[i].argtypes[2] = Iregs;

    fsaves[i].inout[0]  = inarg;
    fsaves[i].inout[1]  = inarg;
    fsaves[i].inout[2]  = inarg;

    fsaves[i].pipe  = FSTRPIPE;
    fsaves[i].code  = i;
  }
  fsaves[CSAVE_INDEXED].argtypes[1]=Iregs;
  fsaves[CS_SAVE_INDEXED].argtypes[1]=Iregs;

  /*Integer arithmetic pipe - Should only need add and subtract */
  /* Both with and without recording */
  adds[IADD].mnemonic = "add  "  ;
  adds[ISUB].mnemonic = "sub  "  ;
  adds[IOR].mnemonic  = "or   "  ;
  adds[CMOVGE].mnemonic = (char *)NULL;
  adds[CMOVGT].mnemonic = "iselgt";
  adds[CMOVLE].mnemonic = "iseleq";
  adds[CMOVLT].mnemonic = "isellt";
  adds[ISUB_PREDICATE].mnemonic = "sub. ";
  adds[IOR_PREDICATE].mnemonic = "or. ";
  adds[IMUL].mnemonic = "mullw ";
  adds[IMUL_IMM].mnemonic = "mulli ";
  adds[IADD_IMM].mnemonic = "addi ";
  adds[IAND_IMM].mnemonic = "andi. ";
  for (i=0;i<nadds;i++){
    adds[i].narg = 3; 
    for(j=0;j<adds[i].narg;j++){
      adds[i].argtypes[j]= Iregs;
      adds[i].inout[j]=inarg;
    }
    adds[i].inout[0] = outarg;
    adds[i].pipe  = IALUPIPE;
    adds[i].code  = i;
  }

  /*Multiply immediate is a special case of IALU OPS*/
  adds[IMUL_IMM].argtypes[2] = SIMM;
  adds[IADD_IMM].argtypes[2] = SIMM;
  adds[IAND_IMM].argtypes[2] = SIMM;

  /*Branch instructions*/
  branch[BRANCH_GT].mnemonic = "bt gt, ";
  branch[BRANCH_EQ].mnemonic = "bt eq, ";
  branch[BRANCH_LT].mnemonic = "bt lt, ";
  branch[BRANCH].mnemonic    = "b      ";
  branch[BRANCH_GE].mnemonic = "bf lt, ";
  branch[BRANCH_LE].mnemonic = "bf gt, ";
  branch[BRANCH_CTR].mnemonic= "bdnz   ";
  branch[BRANCH_RET].mnemonic= "blr    ";
  for (i=0;i<nbrch;i++){
    branch[i].narg = 1;
    branch[i].argtypes[0] = Label;
    branch[i].inout[0]  = inarg;
    branch[i].pipe  = BRCHPIPE;
    branch[i].code  = i;
  }

  /*Assembler directives*/
  directive[Target].mnemonic = "";
  directive[CmovTarget].mnemonic = "";
  directive[Cache_Align].mnemonic = "";
  directive[Enter_Routine].mnemonic = "";
  directive[Exit_Routine].mnemonic = "";
  directive[Pipe_Flush].mnemonic = "";
  directive[NOP].mnemonic = "nop  ";
  directive[FNOP].mnemonic = "fnop ";
  directive[LS_BARRIER].mnemonic = "";

  for (i=0;i<ndirective;i++){
    directive[i].narg = 1;
    directive[i].argtypes[0] = Label;
    directive[i].pipe  = DIRECTIVE;
    directive[i].code  = i;
  }
  weird[WARSTALL].mnemonic = "Write after read";
  weird[LSPIPESPARSE].mnemonic = "L/S pipe busy";
  weird[CACHESTALL].mnemonic = "Stalled by Cache unit";
  for (i=0;i<nweirdops;i++){
    weird[i].narg = 0;
    weird[i].argtypes[0] = Iregs;
    weird[i].argtypes[1] = Iregs;
    weird[i].pipe  = WEIRDRULES;
    weird[i].code  = i;
  }

  /*Now set these up as the instruction set*/
  proc->iset->instructions[SIMDHUMMERPIPE] = hummers;
  proc->iset->instructions[SIMDPERMUTEPIPE]= permutes;
  proc->iset->instructions[FMACPIPE] = fmacs;
  proc->iset->instructions[FADDPIPE] = fadds;
  proc->iset->instructions[FMULPIPE] = fmuls;
  proc->iset->instructions[CACHPIPE] = cache;
  proc->iset->instructions[IALUPIPE] = adds;
  proc->iset->instructions[LOADPIPE] = loads;
  proc->iset->instructions[FLODPIPE] = floads;
  proc->iset->instructions[FSTRPIPE] = fsaves;
  proc->iset->instructions[BRCHPIPE] = branch;
  proc->iset->instructions[DIRECTIVE] = directive;
  proc->iset->instructions[STORPIPE] = stores;
  proc->iset->instructions[FMOVPIPE] = fmovs;
  proc->iset->instructions[WEIRDRULES] = weird;

}

void queue_cmovgt_bgq(int cond,int src,int dst)
{
  make_inst(IALUPIPE,IOR_PREDICATE,cond,cond,cond); /*Sets the predicate reg*/
  make_inst(IALUPIPE,CMOVGT,dst,src,dst);         
}
void queue_cmovlt_bgq(int cond,int src,int dst)
{
  make_inst(IALUPIPE,IOR_PREDICATE,cond,cond,cond); /*Sets the predicate reg*/
  make_inst(IALUPIPE,CMOVLT,dst,src,dst);         
}
void queue_cmovle_bgq(int cond,int src,int dst)
{
  make_inst(IALUPIPE,IOR_PREDICATE,cond,cond,cond); /*Sets the predicate reg*/
  make_inst(IALUPIPE,CMOVLE,dst,src,dst);         
}
void queue_cmovge_bgq(int cond,int src,int dst)
{
  make_inst(IALUPIPE,IOR_PREDICATE,cond,cond,cond); /*Sets the predicate reg*/
  make_inst(IALUPIPE,CMOVLT,dst,src,dst);
}
void cache_print_bgq(struct instruction *i)
{

  switch(i->code){
  case TOUCHL2:
      printf("\t%s 2, \t",i->mnemonic);
      arg_print(i,0);
      printf(",");
      arg_print(i,1);
      printf("\n");
      break;
  case FLUSH:
      printf("\t%s \t",i->mnemonic);
      arg_print(i,0);
      printf(",");
      arg_print(i,1);
      printf(", 3\n");
      break;
  case TOUCHLOCKSET:
  case TOUCHLOCKCLEAR:
  case WRITE_HINT:
  case PREF_IMM:
      printf("\t%s\t",i->mnemonic);
      arg_print(i,0);
      printf(",");
      arg_print(i,1);
      printf("\n");
      break;
  case COUNTER_HINT:
      printf("\t%s\t",i->mnemonic);
      arg_print(i,0);      
      printf("\n");
      break;
  default:
      printf("Error: Unknown cache hint requested\n");
      exit(0);
      break;
  }
  return;
}
