/*
 *
 *  Copyright Peter Boyle and Glasgow University, 2000.
 *  
 *  It is provided under the GNU pubic License V2
 *  It is provided as is and is not guaranteed fit for any purpose.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "processor.h"

/* {...COMPLEX queueing routines */

extern struct processor *PROC;
extern int pregs[];
extern int pimms[];

#define  MAX(a, b)      ((a) > (b) ? (a) : (b))
#define  MIN(a, b)      ((a) < (b) ? (a) : (b))

int have_fmadd;
int fp_latency;
struct rotating_reg *CMADregs=NULL;

int have_hummer(void)
{
  int have_hum = 0;
  int i;
  for ( i =0 ;i < PROC->npipes; i++ ){
    if ( PROC->pipes[i].issuegroups & SIMDHUMMERMASK ) have_hum = 1;
  }
  return have_hum;
}
int have_permute(void)
{
  int have_hum = 0;
  int i;
  for ( i =0 ;i < PROC->npipes; i++ ){
    if ( PROC->pipes[i].issuegroups & SIMDPERMUTEMASK ) have_hum = 1;
  }
  return have_hum;
}
int nsimd(void)
{
  return PROC->nsimd;
}

int reg_tmp = -1;


int i_addr = -1;

int which_is_loaded = -1;

int imm_0 = -1;
int imm_1 = -1;
int imm_2 = -1;
void complex_constants_prepare(int creg, int i_ea)
{
  reg_tmp = creg;
  i_addr  = i_ea;

  if ( imm_0 == -1 ) {
    imm_0 = def_offset(0,Byte,          "simd_zero_offset");
    imm_1 = def_offset(2*nsimd(),Double,"simd_one_offset");
    imm_2 = def_offset(4*nsimd(),Double,"simd_two_offset");
  }
  complex_load(reg_tmp,imm_0,i_addr,Double);
  which_is_loaded = cmplx_i;
}

void complex_constants_assert( int constant)
{
  if ( which_is_loaded == constant ) return;

  if ( constant == cmplx_zero  ) {
    printf("complex_constants_assert: Not supported cmplx_zero");
  }
  fprintf(stderr,"complex_constants_assert : reg_tmp = %d\n",reg_tmp);
  exit(-1);
  make_inst(SIMDHUMMERPIPE,VNEG,reg_tmp,reg_tmp);

  which_is_loaded = constant;
}


void setup_cmadds(void)
{
  int i;
  char *str;
  fp_latency = 0;
  have_fmadd = 0;
  test_proc();

  for ( i =0 ;i < PROC->npipes; i++ ){
    if ( PROC->pipes[i].issuegroups & FMACMASK ) have_fmadd = 1;
    fp_latency = PROC->latency[FMULPIPE];
  } 
  
  if ( !have_fmadd ) {
    CMADregs = create_rotating_reg(Fregs,MAX ( 1+fp_latency,2),"CMADD");
  }
  return;
}

void simd_mul  (int e, int a, int c)
{
  if ( have_hummer() ) { 

    make_inst(SIMDHUMMERPIPE,VMUL,e,a,c);

  } else {

    for (int v=0; v<nsimd() ; v++ ) { 

      int f = e+1; /*RE/IM innermost for hummer style SIMD*/
      int b = a+1;
      int d = c+1;

      queue_fmul(e,a,c);
      queue_fmul(f,b,d);

      e+=2; a+=2; c+=2;
    }

  }
}

void simd_add (int e, int a, int c)
{
  if ( have_hummer() ) { 

    make_inst(SIMDHUMMERPIPE,VADD,e,a,c);

  } else {

    for (int v=0; v<nsimd() ; v++ ) { 

      int f = e+1; /*RE/IM innermost for hummer style SIMD*/
      int b = a+1;
      int d = c+1;

      queue_fadd(e,a,c);
      queue_fadd(f,b,d);

      e+=2; a+=2; c+=2; 
    }
  }
}

void simd_madd (int e, int a, int c,int g)
{
  if ( have_hummer() ) { 

    make_inst(SIMDHUMMERPIPE,VMADD,e,a,c,g);

  } else {

    for (int v=0; v<nsimd() ; v++ ) { 

      int f = e+1; /*RE/IM innermost for hummer style SIMD*/
      int b = a+1;
      int d = c+1;
      int h = g+1;

      queue_fmadd(e,a,c,g);
      queue_fmadd(f,b,d,h);

      e+=2; a+=2; c+=2; g+=2;
    }
  }
}

void simd_nmadd (int e, int a, int c,int g)
{
  if ( have_hummer() ) { 

    make_inst(SIMDHUMMERPIPE,VNMSUB,e,a,c,g);

  } else {

    for (int v=0; v<nsimd() ; v++ ) { 

      int f = e+1; /*RE/IM innermost for hummer style SIMD*/
      int b = a+1;
      int d = c+1;
      int h = g+1;

      queue_fnmsub(e,a,c,g);
      queue_fnmsub(f,b,d,h);

      e+=2; a+=2; c+=2; g+=2;
    }
  }
}


/*Queues (e + if) = (a+ib)*(c+id)*/
void complex_mul(int e,int a,int c)
{ 

  if ( have_hummer() ) { 

    //    complex_constants_assert(cmplx_zero);
    make_inst(SIMDHUMMERPIPE,VMUL_RR_RI,e,a,c,reg_tmp);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e,a,c,e);

  } else {

    for (int v=0; v<nsimd() ; v++ ) { 

      int f = e+1; /*RE/IM innermost for hummer style SIMD*/
      int b = a+1;
      int d = c+1;

      queue_cmul(e,f,a,b,c,d);

      e+=2; a+=2; c+=2;
    }

  }
}
void queue_cmul(int e,int f,int a,int b,int c,int d)
{

  if ( have_fmadd ){
    make_inst(FMULPIPE,FMUL,e,a,c);
    make_inst(FMULPIPE,FMUL,f,a,d);
    make_inst(FMACPIPE,FNMSUB,e,b,d,e);
    make_inst(FMACPIPE,FMADD,f,b,c,f);
  } else{
    int t;

    make_inst(FMULPIPE,FMUL,e,a,c);

    make_inst(FMULPIPE,FMUL,f,a,d);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,d);
    make_inst(FADDPIPE,FSUB,e,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,c);
    make_inst(FADDPIPE,FADD,f,f,t);
  }
  return;
}
/*Queues (e + if) = (e+if) + (a+ib)*(c+id)*/
void complex_madd(int e,int a,int c)
{
  if ( have_hummer() ) { 
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,e,a,c,e);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e,a,c,e);
  } else {

    for (int v=0; v<nsimd() ; v++ ) { 

      int f = e+1; /*RE/IM innermost for hummer style SIMD*/
      int b = a+1;
      int d = c+1;

      queue_cmadd(e,f,a,b,c,d);

      e+=2; a+=2; c+=2;
    }
  }
}

void complex_real_madd(int e, int r, int c)
{

  if ( have_hummer() ) { 
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,e,r,c,e);
  } else {

    for (int v=0; v<nsimd() ; v++ ) { 

      int f = e+1;
      int d = c+1;

      if ( have_fmadd ) { 
        make_inst(FMACPIPE,FMADD,e,r,c,e);
        make_inst(FMACPIPE,FMADD,f,r,d,f);
      } else{
        int t;
        t = get_rotating_register(CMADregs);
        make_inst(FMULPIPE,FMUL,t,r,c);
        make_inst(FADDPIPE,FADD,e,e,t);
        t = get_rotating_register(CMADregs);
        make_inst(FMULPIPE,FMUL,t,r,d);
        make_inst(FADDPIPE,FADD,f,f,t);
      }

      e+=2; c+=2;
    }

  }
}

void complex_real_mul(int e, int r, int c)
{

  if ( have_hummer() ) { 
    make_inst(SIMDHUMMERPIPE,VMUL_RR_RI,e,r,c,e);
  } else {

    for (int v=0; v<nsimd() ; v++ ) { 

      int f = e+1;
      int d = c+1;

      make_inst(FMULPIPE,FMUL,e,r,c);
      make_inst(FMULPIPE,FMUL,f,r,d);

      e+=2; c+=2;
    }

  }
}

void queue_cmadd(int e,int f,int a,int b,int c,int d)
{
  if ( have_fmadd ) {
    make_inst(FMACPIPE,FMADD,e,a,c,e);
    make_inst(FMACPIPE,FMADD,f,a,d,f);
    make_inst(FMACPIPE,FNMSUB,e,b,d,e);
    make_inst(FMACPIPE,FMADD,f,b,c,f);
  }  else {
    int t;

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,c);
    make_inst(FADDPIPE,FADD,e,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,d);
    make_inst(FADDPIPE,FADD,f,f,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,d);
    make_inst(FADDPIPE,FSUB,e,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,c);
    make_inst(FADDPIPE,FADD,f,f,t);

  }
}


/*Queues (h + ig) = (e+if) + (a+ib)*(c+id)*/
void complex_maddto(int h,int e,int a,int c)
{

  if ( have_hummer() ) { 
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,h,a,c,e);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,h,a,c,h);
  } else {
    for (int v=0; v<nsimd() ; v++ ) { 

      int f = e+1;
      int b = a+1;
      int d = c+1;
      int g = h+1;

      queue_cmaddto(h,g,e,f,a,b,c,d);

      e+=2; a+=2; c+=2; h+=2;
    }
  }
}

void queue_cmaddto(int h,int g,int e,int f,int a,int b,int c,int d)
{
  if ( have_fmadd ) {
    make_inst(FMACPIPE,FMADD,h,a,c,e);
    make_inst(FMACPIPE,FMADD,g,a,d,f);
    make_inst(FMACPIPE,FNMSUB,h,b,d,h);
    make_inst(FMACPIPE,FMADD,g,b,c,g);
  } else {
    int t;
    /*Want to leave e and f undisturbed so do the copy in the first one*/
    /*Otherwise risk different behaviour on FMADD and non-FMADD chips*/
    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,c);
    make_inst(FADDPIPE,FADD,h,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,d);
    make_inst(FADDPIPE,FADD,h,f,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,d);
    make_inst(FADDPIPE,FSUB,h,h,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,c);
    make_inst(FADDPIPE,FADD,g,g,t);
  }
}

/*Queues (e + if) = (a-ib)*(c+id)*/
void complex_conjmul(int e,int a,int c)
{
  if ( have_hummer() ) {
    //    complex_constants_assert(cmplx_zero);
    make_inst(SIMDHUMMERPIPE,VMUL_RR_RI,e,a,c,reg_tmp);
    make_inst(SIMDHUMMERPIPE,VMADD_II_MIR,e,a,c,e);
  } else { 

    for (int v=0; v<nsimd() ; v++ ) { 

      int f = e+1;
      int b = a+1;
      int d = c+1;

      queue_conjmul(e,f,a,b,c,d);

      e+=2; a+=2; c+=2;
    }

  }
}
void queue_conjmul(int e,int f,int a,int b,int c,int d)
{

  if ( have_fmadd ) {
    make_inst(FMULPIPE,FMUL,e,a,c);
    make_inst(FMULPIPE,FMUL,f,a,d);
    make_inst(FMACPIPE,FMADD,e,b,d,e);
    make_inst(FMACPIPE,FNMSUB,f,b,c,f);
  }
  else {
    int t;

    make_inst(FMULPIPE,FMUL,e,a,c);

    make_inst(FMULPIPE,FMUL,f,a,d);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,c);
    make_inst(FADDPIPE,FADD,e,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,d);
    make_inst(FADDPIPE,FSUB,f,f,t);
  }
}

void complex_conjmadd(int e,int a,int c)
{
  if ( have_hummer() ) {
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,e,a,c,e);
    make_inst(SIMDHUMMERPIPE,VMADD_II_MIR,e,a,c,e);
  } else { 

    for (int v=0; v<nsimd() ; v++ ) { 

      int f = e+1;
      int b = a+1;
      int d = c+1;

      queue_conjmadd(e,f,a,b,c,d);

      e+=2; a+=2; c+=2; 
    }
  }
}

void queue_conjmadd(int e,int f,int a,int b,int c,int d)
{
  if ( have_fmadd ) {
    make_inst(FMACPIPE,FMADD,e,a,c,e);
    make_inst(FMACPIPE,FMADD,f,a,d,f);
    make_inst(FMACPIPE,FMADD,e,b,d,e);
    make_inst(FMACPIPE,FNMSUB,f,b,c,f);
  }
  else {
    int t;
    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,c);
    make_inst(FADDPIPE,FADD,e,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,d);
    make_inst(FADDPIPE,FADD,f,f,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,d);
    make_inst(FADDPIPE,FADD,e,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,c);
    make_inst(FADDPIPE,FSUB,f,f,t);
  }
}

void complex_conjmaddto(int h, int e,int a,int c)
{
  int f = e+1;
  int b = a+1;
  int d = c+1;
  int g = h+1;

  if ( have_hummer() ) {
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,h,a,c,e);
    make_inst(SIMDHUMMERPIPE,VMADD_II_MIR,h,a,c,h);
  } else {

    for (int v=0; v<nsimd() ; v++ ) { 

      int f = e+1;
      int b = a+1;
      int d = c+1;
      int g = h+1;

      queue_conjmaddto(h,g,e,f,a,b,c,d);

      e+=2; a+=2; c+=2; h+=2;
    }

  }
}

void queue_conjmaddto(int h,int g,int e,int f,int a,int b,int c,int d)
{
  if ( have_fmadd ) {
    make_inst(FMACPIPE,FMADD,h,a,c,e);
    make_inst(FMACPIPE,FMADD,g,a,d,f);
    make_inst(FMACPIPE,FMADD,h,b,d,h);
    make_inst(FMACPIPE,FNMSUB,g,b,c,g);
  }
  else {
    int t;
    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,c);
    make_inst(FADDPIPE,FADD,h,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,d);
    make_inst(FADDPIPE,FADD,g,f,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,d);
    make_inst(FADDPIPE,FADD,h,h,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,c);
    make_inst(FADDPIPE,FSUB,g,g,t);
  }
}


void complex_six_cmuls(int e1,int a1,int c1,
                       int e2,int a2,int c2,
                       int e3,int a3,int c3,
		       int e4,int a4,int c4,
                       int e5,int a5,int c5,
                       int e6,int a6,int c6)
{
  if ( have_hummer() ){

    //complex_constants_assert(cmplx_zero);
    make_inst(SIMDHUMMERPIPE,VMUL_RR_RI,e1,a1,c1,reg_tmp);
    make_inst(SIMDHUMMERPIPE,VMUL_RR_RI,e2,a2,c2,reg_tmp);
    make_inst(SIMDHUMMERPIPE,VMUL_RR_RI,e3,a3,c3,reg_tmp);
    make_inst(SIMDHUMMERPIPE,VMUL_RR_RI,e4,a4,c4,reg_tmp);
    make_inst(SIMDHUMMERPIPE,VMUL_RR_RI,e5,a5,c5,reg_tmp);
    make_inst(SIMDHUMMERPIPE,VMUL_RR_RI,e6,a6,c6,reg_tmp);

    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e1,a1,c1,e1);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e2,a2,c2,e2);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e3,a3,c3,e3);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e4,a4,c4,e4);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e5,a5,c5,e5);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e6,a6,c6,e6);

  } else { 
    complex_three_cmuls(e1,a1,c1,
			e2,a2,c2,
			e3,a3,c3);
    complex_three_cmuls(e4,a4,c4,
			e5,a5,c5,
			e6,a6,c6);
  }

}
void complex_six_cmadds(int e1,int a1,int c1,
                       int e2,int a2,int c2,
                       int e3,int a3,int c3,
		       int e4,int a4,int c4,
                       int e5,int a5,int c5,
                       int e6,int a6,int c6)
{
  if ( have_hummer() ){

    //complex_constants_assert(cmplx_zero);
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,e1,a1,c1,e1);
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,e2,a2,c2,e2);
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,e3,a3,c3,e3);
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,e4,a4,c4,e4);
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,e5,a5,c5,e5);
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,e6,a6,c6,e6);

    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e1,a1,c1,e1);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e2,a2,c2,e2);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e3,a3,c3,e3);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e4,a4,c4,e4);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e5,a5,c5,e5);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e6,a6,c6,e6);

  } else { 
    complex_three_cmadds(e1,a1,c1,
			 e2,a2,c2,
			 e3,a3,c3);
    complex_three_cmadds(e4,a4,c4,
			 e5,a5,c5,
			 e6,a6,c6);
  }
}

void complex_three_cmuls(int e1,int a1,int c1,
                       int e2,int a2,int c2,
                       int e3,int a3,int c3)
{

  if ( have_hummer() ){

    //complex_constants_assert(cmplx_zero);
    make_inst(SIMDHUMMERPIPE,VMUL_RR_RI,e1,a1,c1,reg_tmp);
    make_inst(SIMDHUMMERPIPE,VMUL_RR_RI,e2,a2,c2,reg_tmp);
    make_inst(SIMDHUMMERPIPE,VMUL_RR_RI,e3,a3,c3,reg_tmp);

    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e1,a1,c1,e1);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e2,a2,c2,e2);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e3,a3,c3,e3);

/*
 * } else if ( have_fmadd ){
 *
 * int f1 = e1 + 1;
 * int f2 = e2 + 1;
 * int f3 = e3 + 1;

 * int b1 = a1 + 1;
 * int b2 = a2 + 1;
 * int b3 = a3 + 1;

 * int d1 = c1 + 1;
 * int d2 = c2 + 1;
 *  int d3 = c3 + 1;
 *    queue_three_cmuls(e1, f1, a1, b1, c1, d1,
 *		      e2, f2, a2, b2, c2, d2,
 *		      e3, f3, a3, b3, c3, d3);
 */
  } else{
    complex_mul(e1,a1,c1);
    complex_mul(e2,a2,c2);
    complex_mul(e3,a3,c3);
  }
  return;
}
void queue_three_cmuls(int e1, int f1, int a1, int b1, int c1, int d1,
		       int e2, int f2, int a2, int b2, int c2, int d2,
		       int e3, int f3, int a3, int b3, int c3, int d3)
{ 
  queue_cmul(e1,f1,a1,b1,c1,d1);
  queue_cmul(e2,f2,a2,b2,c2,d2);
  queue_cmul(e3,f3,a3,b3,c3,d3);
}

void complex_three_cmadds(int e1,int a1,int c1,
                        int e2,int a2,int c2,
                        int e3,int a3,int c3)
{
  if ( have_hummer() ){

    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,e1,a1,c1,e1);
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,e2,a2,c2,e2);
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,e3,a3,c3,e3);

    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e1,a1,c1,e1);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e2,a2,c2,e2);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e3,a3,c3,e3);

  } else {
    complex_madd(e1,a1,c1);
    complex_madd(e2,a2,c2);
    complex_madd(e3,a3,c3);
  }
  return;
}

void queue_three_cmadds(int e1, int f1, int a1, int b1, int c1, int d1,
		       int e2, int f2, int a2, int b2, int c2, int d2,
		       int e3, int f3, int a3, int b3, int c3, int d3)
{
  if ( have_fmadd ) {

    make_inst(FMACPIPE,FMADD,e1,a1,c1,e1);
    make_inst(FMACPIPE,FMADD,f1,a1,d1,f1);

    make_inst(FMACPIPE,FMADD,e2,a2,c2,e2);
    make_inst(FMACPIPE,FMADD,f2,a2,d2,f2);

    make_inst(FMACPIPE,FMADD,e3,a3,c3,e3);
    make_inst(FMACPIPE,FMADD,f3,a3,d3,f3);

    make_inst(FMACPIPE,FNMSUB,e1,b1,d1,e1);
    make_inst(FMACPIPE,FMADD,f1,b1,c1,f1);

    make_inst(FMACPIPE,FNMSUB,e2,b2,d2,e2);
    make_inst(FMACPIPE,FMADD,f2,b2,c2,f2);

    make_inst(FMACPIPE,FNMSUB,e3,b3,d3,e3);
    make_inst(FMACPIPE,FMADD,f3,b3,c3,f3);

  } else { 

    queue_cmadd(e1,f1,a1,b1,c1,d1);
    queue_cmadd(e2,f2,a2,b2,c2,d2);
    queue_cmadd(e3,f3,a3,b3,c3,d3);

  }
}

void complex_three_cmaddtos(
                       int h1,int e1,int a1,int c1,
                       int h2,int e2,int a2,int c2,
                       int h3,int e3,int a3,int c3)
{
  if ( have_hummer() ){

    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,h1,a1,c1,e1);
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,h2,a2,c2,e2);
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,h3,a3,c3,e3);

    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,h1,a1,c1,h1);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,h2,a2,c2,h2);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,h3,a3,c3,h3);

  } else {
    complex_maddto(h1,e1,a1,c1);
    complex_maddto(h2,e2,a2,c2);
    complex_maddto(h3,e3,a3,c3);
  }
  return;
}
void queue_three_cmaddtos(int h1,int g1,int e1,int f1,int a1,int b1,int c1,int d1,
		     int h2,int g2,int e2,int f2,int a2,int b2,int c2,int d2,
		     int h3,int g3,int e3,int f3,int a3,int b3,int c3,int d3)
{
  if ( have_fmadd ) {
    
    make_inst(FMACPIPE,FMADD,h1,a1,c1,e1);
    make_inst(FMACPIPE,FMADD,g1,a1,d1,f1);

    make_inst(FMACPIPE,FMADD,h2,a2,c2,e2);
    make_inst(FMACPIPE,FMADD,g2,a2,d2,f2);

    make_inst(FMACPIPE,FMADD,h3,a3,c3,e3);
    make_inst(FMACPIPE,FMADD,g3,a3,d3,f3);

    make_inst(FMACPIPE,FNMSUB,h1,b1,d1,h1);
    make_inst(FMACPIPE,FMADD,g1,b1,c1,g1);

    make_inst(FMACPIPE,FNMSUB,h2,b2,d2,h2);
    make_inst(FMACPIPE,FMADD,g2,b2,c2,g2);

    make_inst(FMACPIPE,FNMSUB,h3,b3,d3,h3);
    make_inst(FMACPIPE,FMADD,g3,b3,c3,g3);

  } else {
    queue_cmaddto(h1,g1,e1,f1,a1,b1,c1,d1);
    queue_cmaddto(h2,g2,e2,f2,a2,b2,c2,d2);
    queue_cmaddto(h3,g3,e3,f3,a3,b3,c3,d3);
  }
} 

void complex_three_conjmadds(int e1,int a1,int c1,
			     int e2,int a2,int c2,
			     int e3,int a3,int c3)
{
  if ( have_hummer() ){

    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,e1,a1,c1,e1);
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,e2,a2,c2,e2);
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,e3,a3,c3,e3);

    make_inst(SIMDHUMMERPIPE,VMADD_II_MIR,e1,a1,c1,e1);
    make_inst(SIMDHUMMERPIPE,VMADD_II_MIR,e2,a2,c2,e2);
    make_inst(SIMDHUMMERPIPE,VMADD_II_MIR,e3,a3,c3,e3);

  } else{
    complex_conjmadd(e1,a1,c1);
    complex_conjmadd(e2,a2,c2);
    complex_conjmadd(e3,a3,c3);
  }
  return;

}

void queue_three_conjmadds(int e1,int f1,int a1,int b1,int c1,int d1,
			   int e2,int f2,int a2,int b2,int c2,int d2,
			   int e3,int f3,int a3,int b3,int c3,int d3)
{
  if ( have_fmadd ) {
    
    make_inst(FMACPIPE,FMADD,e1,a1,c1,e1);
    make_inst(FMACPIPE,FMADD,f1,a1,d1,f1);

    make_inst(FMACPIPE,FMADD,e2,a2,c2,e2);
    make_inst(FMACPIPE,FMADD,f2,a2,d2,f2);

    make_inst(FMACPIPE,FMADD,e3,a3,c3,e3);
    make_inst(FMACPIPE,FMADD,f3,a3,d3,f3);

    make_inst(FMACPIPE,FMADD,e1,b1,d1,e1);
    make_inst(FMACPIPE,FNMSUB,f1,b1,c1,f1);

    make_inst(FMACPIPE,FMADD,e2,b2,d2,e2);
    make_inst(FMACPIPE,FNMSUB,f2,b2,c2,f2);

    make_inst(FMACPIPE,FMADD,e3,b3,d3,e3);
    make_inst(FMACPIPE,FNMSUB,f3,b3,c3,f3);

  } else {

     queue_conjmadd(e1,f1,a1,b1,c1,d1);
     queue_conjmadd(e2,f2,a2,b2,c2,d2);
     queue_conjmadd(e3,f3,a3,b3,c3,d3);


  }
}


void complex_three_conjmaddtos(
                       int h1,int e1,int a1,int c1,
                       int h2,int e2,int a2,int c2,
                       int h3,int e3,int a3,int c3)
{
  if ( have_hummer() ){

    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,h1,a1,c1,e1);
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,h2,a2,c2,e2);
    make_inst(SIMDHUMMERPIPE,VMADD_RR_RI,h3,a3,c3,e3);

    make_inst(SIMDHUMMERPIPE,VMADD_II_MIR,h1,a1,c1,h1);
    make_inst(SIMDHUMMERPIPE,VMADD_II_MIR,h2,a2,c2,h2);
    make_inst(SIMDHUMMERPIPE,VMADD_II_MIR,h3,a3,c3,h3);

  } else{

    complex_conjmaddto(h1,e1,a1,c1);
    complex_conjmaddto(h2,e2,a2,c2);
    complex_conjmaddto(h3,e3,a3,c3);

  }
  return;
}

void queue_three_conjmaddtos(int h1,int g1,int e1,int f1,int a1,int b1,int c1,int d1,
			     int h2,int g2,int e2,int f2,int a2,int b2,int c2,int d2,
			     int h3,int g3,int e3,int f3,int a3,int b3,int c3,int d3)
{
  if ( have_fmadd ) { 
    
    make_inst(FMACPIPE,FMADD,h1,a1,c1,e1);
    make_inst(FMACPIPE,FMADD,g1,a1,d1,f1);

    make_inst(FMACPIPE,FMADD,h2,a2,c2,e2);
    make_inst(FMACPIPE,FMADD,g2,a2,d2,f2);

    make_inst(FMACPIPE,FMADD,h3,a3,c3,e3);
    make_inst(FMACPIPE,FMADD,g3,a3,d3,f3);

    make_inst(FMACPIPE,FMADD,h1,b1,d1,h1);
    make_inst(FMACPIPE,FNMSUB,g1,b1,c1,g1);

    make_inst(FMACPIPE,FMADD,h2,b2,d2,h2);
    make_inst(FMACPIPE,FNMSUB,g2,b2,c2,g2);

    make_inst(FMACPIPE,FMADD,h3,b3,d3,h3);
    make_inst(FMACPIPE,FNMSUB,g3,b3,c3,g3);

  } else {

    queue_conjmaddto(h1,g1,e1,f1,a1,b1,c1,d1);
    queue_conjmaddto(h2,g2,e2,f2,a2,b2,c2,d2);
    queue_conjmaddto(h3,g3,e3,f3,a3,b3,c3,d3);
    
  }
}

void complex_three_conjmuls(int e1,int a1,int c1,
			    int e2,int a2,int c2,
			    int e3,int a3,int c3)
{
  if ( have_hummer() ){

    make_inst(SIMDHUMMERPIPE,VMUL_RR_RI,e1,a1,c1);
    make_inst(SIMDHUMMERPIPE,VMUL_RR_RI,e2,a2,c2);
    make_inst(SIMDHUMMERPIPE,VMUL_RR_RI,e3,a3,c3);

    make_inst(SIMDHUMMERPIPE,VMADD_II_MIR,e1,a1,c1,e1);
    make_inst(SIMDHUMMERPIPE,VMADD_II_MIR,e2,a2,c2,e2);
    make_inst(SIMDHUMMERPIPE,VMADD_II_MIR,e3,a3,c3,e3);

  } else{
    complex_conjmul(e1,a1,c1);
    complex_conjmul(e2,a2,c2);
    complex_conjmul(e3,a3,c3);
  }
  return;

}

void queue_three_conjmuls(int e1,int f1,int a1,int b1,int c1,int d1,
			  int e2,int f2,int a2,int b2,int c2,int d2,
			  int e3,int f3,int a3,int b3,int c3,int d3)
{

  if ( have_fmadd ) {

    make_inst(FMULPIPE,FMUL,e1,a1,c1);
    make_inst(FMULPIPE,FMUL,f1,a1,d1);

    make_inst(FMULPIPE,FMUL,e2,a2,c2);
    make_inst(FMULPIPE,FMUL,f2,a2,d2);

    make_inst(FMULPIPE,FMUL,e3,a3,c3);
    make_inst(FMULPIPE,FMUL,f3,a3,d3);

    make_inst(FMACPIPE,FMADD,e1,b1,d1,e1);
    make_inst(FMACPIPE,FNMSUB,f1,b1,c1,f1);

    make_inst(FMACPIPE,FMADD,e2,b2,d2,e2);
    make_inst(FMACPIPE,FNMSUB,f2,b2,c2,f2);

    make_inst(FMACPIPE,FMADD,e3,b3,d3,e3);
    make_inst(FMACPIPE,FNMSUB,f3,b3,c3,f3);

  } else { 

    queue_conjmul(e1,f1,a1,b1,c1,d1);
    queue_conjmul(e2,f2,a2,b2,c2,d2);
    queue_conjmul(e3,f3,a3,b3,c3,d3);

  }
}

void queue_fmac ( int d, int a , int b ) 
{
  queue_fmadd(d,a,b,d);
}
void queue_fnmac ( int d, int a , int b ) 
{
  queue_fnmsub(d,a,b,d);
}

void queue_fnmsub ( int d, int a , int b, int c) 
{
  int t;
  if ( have_fmadd ){
    make_inst(FMACPIPE,FNMSUB,d,a,b,c);
  }
  else {
    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,b);
    make_inst(FADDPIPE,FSUB,d,c,t);
  }
}
void queue_fmadd ( int d, int a , int b, int c ) 
{
  int t;
  if ( have_fmadd ){
    make_inst(FMACPIPE,FMADD,d,a,b,c);
  }
  else {
    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,b);
    make_inst(FADDPIPE,FADD,d,c,t);
  }
}


/*Bugger is there are no immediate offsets on the damn hummer*/
/*this forces endless loading*/
void complex_load     (int dest,int Himm,int basereg, enum Datum t)
{
  int index_reg;
  if ( have_hummer() ) {
    index_reg = pregs[get_constant(get_offset(Himm))];
    if ( t == Double ) 
      make_inst(FLODPIPE,CLOAD_INDEXED,dest,index_reg,basereg);
    else 
      make_inst(FLODPIPE,CS_LOAD_INDEXED,dest,index_reg,basereg);

  } else { 

    for(int v=0;v<nsimd()*2;v++){

      int destv = dest+v;

      int units = get_offset(Himm)+ v*SizeofDatum(t);
      int Himmv = get_offset_handle(units,Byte);
      queue_fload(destv,Himmv,basereg,t);

    }      

  }
}
void complex_store    (int src ,int Himm,int basereg,enum Datum t)
{
  int index_reg;
  if ( have_hummer() ) {
    index_reg = pregs[get_constant(get_offset(Himm))];
    if ( t == Double ) 
      make_inst(FSTRPIPE,CSAVE_INDEXED,src,index_reg,basereg);
    else 
      make_inst(FSTRPIPE,CS_SAVE_INDEXED,src,index_reg,basereg);
  } else { 

    for(int v=0;v<nsimd()*2;v++){

      int srcv = src+v;
      int units = get_offset(Himm)+ v*SizeofDatum(t);
      int Himmv = get_offset_handle(units,Byte);
      queue_fstore(srcv,Himmv,basereg,t);
 
    }
  }
}
void complex_add (int e, int a, int c)
{
  int b = a+1;
  int f = e+1;
  int d = c+1;

  if ( have_hummer() ) {
    make_inst(SIMDHUMMERPIPE,VADD,e,a,c);
  } else { 
    for(int v=0;v<2*nsimd();v++) {
      queue_fadd(e+v,a+v,c+v);
    }
  }
}
void complex_sub (int e, int a, int c)
{
  int b = a+1;
  int f = e+1;
  int d = c+1;

  if ( have_hummer() ) {
    make_inst(SIMDHUMMERPIPE,VSUB,e,a,c);
  } else { 
    for(int v=0;v<2*nsimd();v++) {
      queue_fsub(e+v,a+v,c+v);
    }
  }
}

void complex_ApiB (int e, int a, int c)
{

  if ( have_hummer() ) {
    complex_constants_assert(cmplx_i);
    make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e,reg_tmp,c,a);
  } else { 
    for(int v=0;v<nsimd();v++) {

      int b = a+1, f = e+1, d = c+1;
      queue_fsub(e,a,d);
      queue_fadd(f,b,c);
      a+=2; e+=2; c+=2;
    }
  }
}
void complex_AmiB (int e, int a, int c)
{
  int b = a+1;
  int f = e+1;
  int d = c+1;

  if ( have_hummer() ) {
    if ( PROC->cross_negative ) {
      complex_constants_assert(cmplx_i);
      make_inst(SIMDHUMMERPIPE,VMADD_II_MIR,e,reg_tmp,c,a);
      //      make_inst(SIMDHUMMERPIPE,VMADD_II_MIR,e,c,reg_tmp,a);
    } else {
      complex_constants_assert(cmplx_neg_i);
      make_inst(SIMDHUMMERPIPE,VMADD_MII_IR,e,reg_tmp,c,a);
    }
  } else { 
    for(int v=0;v<nsimd();v++) {
      int b = a+1, f = e+1, d = c+1;
      queue_fadd(e,a,d);
      queue_fsub(f,b,c);
      a+=2; e+=2; c+=2;
    }    
  }
}
/* ...END complex queueing routines*/


/*
 * Provided as general primitives for treating SIMD index
 * as a 2^n grid. I need to clarify my thoughts to code this
 * for general vector length.
 */
static int perm  = (0x2<<9)|(0x3<<6)|(0x0<<3)|0x1 ;
static int lolo  = (0x0<<9)|(0x1<<6)|(0x4<<3)|0x5 ;
static int hihi  = (0x2<<9)|(0x3<<6)|(0x6<<3)|0x7 ;
static int lohi  = (0x0<<9)|(0x1<<6)|(0x6<<3)|0x7 ;
static int permreg = -1;
static int permval = -1;
void complex_simd_init(int preg)
{
  perm  = def_offset((0x2<<9)|(0x3<<6)|(0x0<<3)|0x1,Byte,"perm") ; //2,3,0,1
  lolo  = def_offset((0x0<<9)|(0x1<<6)|(0x4<<3)|0x5,Byte,"lolo") ; //0,1,4,5
  hihi  = def_offset((0x2<<9)|(0x3<<6)|(0x6<<3)|0x7,Byte,"hihi") ; //2,4,6,7
  lohi  = def_offset((0x0<<9)|(0x1<<6)|(0x6<<3)|0x7,Byte,"lohi") ; //0,1,6,7
  permreg = preg;
}
void complex_simd_permute (int mu, int dest, int src)
{ 
  if ( nsimd() != 2 ) exit(0);
  if ( mu != 3 ) exit(0);
 
  complex_simd_permute_internal(dest, src, src, perm) ;
}
void complex_simd_extract (int hi, int mu, int dest , int src1 , int src2)
{
  if ( nsimd() != 2 ) exit(0);
  if ( mu != 3 ) exit(0);

  if ( hi ) complex_simd_permute_internal(dest, src1, src2, hihi) ;
  else      complex_simd_permute_internal(dest, src1, src2, lolo) ;
}
void complex_simd_merge   (int hi, int mu, int dest1, int src1, int src2 )
{
  if ( nsimd() != 2 ) exit(0);
  if ( mu != 3 ) exit(0);
  
  if ( hi ) {
    complex_simd_permute_internal(dest1, src1, src2, hihi) ;
  } else { 
    complex_simd_permute_internal(dest1, src1, src2, lolo) ;
  }
}

void complex_simd_permute_internal (int dest, int src1, int src2, int hperm)
{
  int perm = get_offset(hperm);
  if ( have_permute() ) { 
    if ( permval != perm ) {
      make_inst(SIMDPERMUTEPIPE,VPERMIMM,permreg,hperm);
      permval = perm;
    }
    make_inst(SIMDPERMUTEPIPE,VPERM,dest,src1,src2,permreg);
  } else { 
    for (int i=0;i<4; i++ ) { 
      int pi = (perm>>(3*(3-i)))&0x7;
      if ( pi < 4 ) queue_fmov(permreg+i,src1+pi);
      else          queue_fmov(permreg+i,src2+pi-4);
    }
    for (int i=0;i<4; i++ ) { 
      queue_fmov(dest+i,permreg+i);
    }    
  }
}

