/*
 *
 *  Copyright Peter Boyle and Glasgow University, 2000.
 *  
 *  It is provided under the GNU pubic License V2
 *  It is provided as is and is not guaranteed fit for any purpose.
 *
 */

#include "processor.h"

extern struct processor *PROC;
int N_Ahead = 1;
int frame_down;
int frame_up;
int frame_regbase;
int frame_whbase;/*Prefetch bit bucket*/
int frame_membase;
/*
 * Some cpu's have rules about the stack pointer being aligned on a
 * 16 byte boundary. (US-II) If we ever support something with other
 * alignment constraints, easiest to just use lcm as a unified constraint.
 * Return offset from stack pointer for *LOW* end of allocated buffer
 */
#define STACK_ALIGN 16

/*
 * This routine returns a byte offset handle to a bias from the stack pointer
 * that gives the *low* end of the alloced space
 *
 * The registers are saved in a space *above* the allocated memory
 * If REGWINDOWS are used we cannot allocate stack portably, and raise an error
 * If no regwindows are used then we can only produce leaf routines on US-II
 * 
 * We guarantee it is safe to both prefetch and write hint to the stack pointer
 *
 */
int grab_stack ( int nbyte )
{
  int unitsize = PROC->FP_size;
  int frame;
  int bias;
  int alignbyte;   

  if (PROC->id >= UNKNOWN ) {
    def_off ( bias , Byte, 0) ;
    /*    fprintf(stderr,"nbyte =%d\n",nbyte );*/
    frame_membase = nbyte;
    return bias;
  }

  /*Make sure we only align frames on (16 byte) bdy's*/

  alignbyte = STACK_ALIGN* ( ((nbyte+STACK_ALIGN-1)/STACK_ALIGN) );
  if ( alignbyte != nbyte ) {
    fprintf(stderr, "grab_stack_alloc\n");
    fprintf(stderr, "Warning aligning stack frame: %d bytes requested\n",nbyte);
  }


  /*
   * We could redefine the locations of mem and regs to match the reg windows
   * on US-II,III
   * Haven't done as yet, and this restricts us to leaf routines
   * on US...
   */

  /*Skip Bit bucket for prefetching*/
  frame_regbase = 0 ;
  frame_whbase = (PROC->iregs->regfile+PROC->fregs->regfile)*8 
                +((N_Ahead+2)*PROC->CacheLine);
  frame_membase  = frame_whbase + ((N_Ahead+3)*PROC->CacheLine);
  frame = frame_membase + nbyte;
  /*Skip Bit bucket and data to regs*/

  def_off( frame_down_tmp, Byte, -frame ); frame_down= frame_down_tmp;
  def_off( frame_up_tmp, Byte, frame); frame_up = frame_up_tmp;

  /*SPARC v9 has an annoying 2047 implicit offset from sp, to stack addr   */
  /* this is called bias in documentation. Hence PROC->Bias member         */
  /* I corrupt this terminology  slightly to                               */
  /* mean a bias to sp that points to the data storage on stack for routine*/
  bias = def_offset ( PROC->Bias+frame_membase, Byte, "bias" );

  if ( PROC->RegWindows ) {
    make_inst(DIRECTIVE,PUSH_WINDOW,frame_down);
    return(bias);
  }else {
    queue_load_addr(PROC->StackPointer,frame_down,PROC->StackPointer);
    return(bias);    
  }
}

void free_stack ( void )
{
  int unitsize = PROC->FP_size;

  if (PROC->id >= UNKNOWN ) return;

  if ( PROC->RegWindows ) {
    make_inst(DIRECTIVE,POP_WINDOW,frame_up);
  }
  else{
    queue_load_addr(PROC->StackPointer,frame_up,PROC->StackPointer);
  }
}

int SAVE_I[32];
int SAVE_F[32];
void save_regs(void)
{
  int i;
  char *str;
  int unitsize = PROC->FP_size;
  
  if (PROC->id >= UNKNOWN ) return;

  if ( !PROC->RegWindows ){
    for ( i = 0 ; i <PROC->iregs->regfile ; i ++ ){
      if ( PROC->Isave & (1<<i) ) {
        str=(char *)malloc(10);
        sprintf(str,"S_I_%d",i);

	/*Add in PROC->Bias for US-II in v9 mode*/

        SAVE_I[i] = 
         def_offset( frame_regbase+(PROC->fregs->regfile)*8
                          +PROC->Bias + i*8,Byte ,str );
        make_inst(STORPIPE,SAVE_REG_EA,i,SAVE_I[i],PROC->StackPointer);
      }
    }
    for ( i = 0 ; i <PROC->fregs->regfile ; i ++ ){
      if ( PROC->Fsave & (1<<i) ) {
        str=(char *)malloc(10);
        sprintf(str,"S_F_%d",i);

	/*Add in PROC->Bias for US-II in v9 mode*/

        SAVE_F[i] = 
          def_offset( frame_regbase + i*8
                          +PROC->Bias, Byte, str );
        make_inst(FSTRPIPE,FSAVE_REG_EA,i,SAVE_F[i],PROC->StackPointer);
     }
    }
  }
  return;
}
void restore_regs(void)
{
  int i;
  char *str;

  if (PROC->id >= UNKNOWN ) return;

  if ( ! PROC->RegWindows ){
    for ( i = 0 ; i <PROC->iregs->regfile ; i ++ ){
      if ( PROC->Isave & (1<<i) ) {
        make_inst(LOADPIPE,LOAD_REG_EA,i,SAVE_I[i],PROC->StackPointer);
      }
    }
    for ( i = 0 ; i <PROC->fregs->regfile ; i ++ ){
      if ( PROC->Fsave & (1<<i) ) {
        make_inst(FLODPIPE,FLOAD_REG_EA,i,SAVE_F[i],PROC->StackPointer);
      }
    }
  }
  return;
}
